#include "oss_config.h"

#include <stdio.h>
#include <sys/stat.h>


#define MAX_CARDS 16


extern int wait_for_interrupt(void);
extern void cancel_timer(timer_t *timer);
extern void set_timer(timer_t *timer, clock_t ticks, tmr_func_t function, void *arg);




int oss_num_cards = 0;
static oss_device_t *cards[MAX_CARDS];

static struct fileinfo files[/*OSS_MAX_CDEVS*/256];

int oss_major = 13;
static int open_devices = 0;
static oss_device_t core_osdev;

static int oss_expired = 0;



typedef struct tmout_desc {
	volatile int active;
	int timestamp;
	void (*func) (void *);
	void *arg;

	timer_t timer;
} tmout_desc_t;

static volatile int next_id = 0;
#define MAX_TMOUTS 128

tmout_desc_t tmouts[MAX_TMOUTS] = { { 0 } };

int timeout_random = 0x12123400;





	static int
grow_array(oss_device_t *osdev, oss_cdev_t ***arr, int *size, int element_size, int increment)
{
	oss_cdev_t **old=*arr, **new = *arr;
	int old_size = *size;
	int new_size = *size;

	new_size += increment;

	if ((new=PMALLOC(osdev, new_size * element_size))==NULL)
		return 0;

	memset(new, 0, new_size * element_size);
	if (old != NULL)
		memcpy(new, old, old_size * element_size);

	*size = new_size;
	*arr = new;

	if (old != NULL)
		PMFREE(osdev, old);

	return 1;
}

void oss_install_chrdev(oss_device_t *osdev, char *name, int dev_class, 
		int instance, oss_cdev_drv_t *drv, int flags)
{
	static char fname[50];
	oss_cdev_t *cdev = NULL;
	int num;
	FILE* f;

	if (dev_class != OSS_DEV_STATUS)
		if (oss_expired && instance > 0)
			return;

	for (num = 0; num < oss_num_cdevs; num++)
		if (oss_cdevs[num]->d == NULL)  /* Unloaded driver */
			if (oss_cdevs[num]->dev_class == dev_class
					&& oss_cdevs[num]->instance == instance)
			{
				cdev = oss_cdevs[num];
				break;
			}
	if (cdev == NULL)
	{
		if (oss_num_cdevs >= OSS_MAX_CDEVS)
		{
			if (!grow_array(osdev, &oss_cdevs, &oss_max_cdevs, sizeof(oss_cdev_t*), 100))
			{
				cmn_err (CE_WARN, "Cannot allocate new minor numbers.\n");
				return;
			}
		}

		if ((cdev = PMALLOC (NULL, sizeof (*cdev))) == NULL)
		{
			cmn_err (CE_WARN, "Cannot allocate character device desc.\n");
			return;
		}
		num = oss_num_cdevs++;
	}
	memset (cdev, 0, sizeof (*cdev));
	cdev->dev_class = dev_class;
	cdev->instance = instance;
	cdev->d = drv;
	cdev->osdev = osdev;
	if (name != NULL)
		strncpy (cdev->name, name, sizeof (cdev->name) - 1);
	else
		strcpy (cdev->name, "NONE");
	cdev->name[sizeof (cdev->name) - 1] = 0;
	oss_cdevs[num] = cdev;

	snprintf(fname, sizeof(fname), "/dev/%s", cdev->name);
	unlink(fname);
	if(mknod(fname, S_IFCHR | 0666, makedev(core_osdev.major, num)) < 0) {
		printf("OSS: couldn't mknod %s\n", fname);
	}
}


void *oss_get_osid(oss_device_t *osdev) {
	return osdev->osid;
}


void oss_remove_wait_queue(oss_wait_queue_t *wq) {
	cmn_err(0, "Not implemented: oss_remove_wait_queue()\n");
}


oss_wait_queue_t *oss_create_wait_queue(oss_device_t *osdev, const char *name) {
	oss_wait_queue_t *wq;

	if ((wq = PMALLOC(NULL, sizeof(*wq))) == NULL)
		return NULL;

	wq->wakeup = 0;
	wq->expired = 0;
	tmr_inittimer(&wq->timer);

	return wq;
}


void oss_reset_wait_queue(oss_wait_queue_t *wq) {
	wq->flags = 0;
	wq->wakeup = 0;
	wq->expired = 0;
}


int __oss_alloc_dmabuf(int dev, dmap_p dmap, unsigned int alloc_flags,
		oss_uint64_t maxaddr, int direction) {
	void *tmpbuf;
	oss_native_word phaddr;
	int size = 64 * 1024;
	extern int dma_buffsize;

	if (dma_buffsize > 16 && dma_buffsize <= 128)
		size = dma_buffsize * 1024;

	if (dmap->dmabuf != NULL)
		return 0;

	/*
	 * Some applications and virtual drivers need shorter buffer.
	 */
	if (dmap->flags & DMAP_SMALLBUF)
	{
		size = SMALL_DMABUF_SIZE;
	}
	else if (dmap->flags & DMAP_MEDIUMBUF)
	{
		size = MEDIUM_DMABUF_SIZE;
	}

	if ((alloc_flags & DMABUF_SIZE_16BITS) && size > 32 * 1024)
		size = 32 * 1024;

	tmpbuf = CONTIG_MALLOC (dmap->osdev, size, maxaddr, &phaddr, NULL);
	if (tmpbuf == NULL)
		return OSS_ENOMEM;
	dmap->dmabuf = tmpbuf;
	dmap->buffsize = size;
	dmap->dmabuf_phys = phaddr;

	return 0;
}


void oss_wakeup(oss_wait_queue_t *wq, oss_mutex_t *mutex, 
		oss_native_word *flags, short events) {
	wq->wakeup = 1;
}


void oss_free_dmabuf(int dev, dmap_p dmap) {
	if (dmap->dmabuf == NULL) {
		return;
	}

	CONTIG_FREE (dmap->osdev, dmap->dmabuf, dmap->buffsize, NULL);
	dmap->dmabuf = NULL;
	dmap->buffsize = 0;
	dmap->dmabuf_phys = 0;
}


unsigned long oss_get_time(void) {
	clock_t clock_ticks = 0;

	if (getuptime(&clock_ticks) < 0) {
		cmn_err(CE_PANIC, "Cannot get uptime.\n");
	}
	return (unsigned long) clock_ticks;
}   


void set_wq_expired(struct timer *tp) {
	oss_wait_queue_t *wq = (oss_wait_queue_t *) tmr_arg(tp)->ta_ptr;
	wq->expired = 1;
}


int oss_sleep(oss_wait_queue_t *wq, oss_mutex_t *mutex, int ticks, 
		oss_native_word *flags, unsigned int *status) {
	int result_waiting;

	*status = 0;

	set_timer(&wq->timer, (clock_t) ticks, set_wq_expired, wq);

	while(1) {
		if (wq->wakeup) {
			wq->wakeup = 0;
			wq->expired = 0;
			cancel_timer(&wq->timer);
			return 1;
		}
		else if (wq->expired) {
			wq->wakeup = 0;
			wq->expired = 0;
			return 0;
		}

		result_waiting = wait_for_interrupt();

		if (result_waiting == -1) { /* caught a signal */
			*status |= WK_SIGNAL; 
			wq->wakeup = 0;
			cancel_timer(&wq->timer);
			return 1;
		}
		else if (result_waiting == -2) { /* message_queue is full */
			cmn_err(0, "information: message_queue full.\n");
			cancel_timer(&wq->timer);
			return 0;
		}
		else {
			/* just continue waiting for interrupts */
		}
	}
	cancel_timer(&wq->timer);
	return 0;
}


int oss_get_cardinfo(int cardnum, oss_card_info *ci) {
	if (cardnum < 0 || cardnum >= oss_num_cards) {
		return OSS_ENXIO;
	}

	if (cards[cardnum]->name != NULL) {
		strncpy (ci->longname, cards[cardnum]->name, 128);
	}
	ci->shortname[127] = 0;

	if (cards[cardnum]->nick != NULL) {
		strncpy (ci->shortname, cards[cardnum]->nick, 16);
	}
	ci->shortname[15] = 0;

	if (cards[cardnum]->hw_info != NULL) {
		strncpy (ci->hw_info, cards[cardnum]->hw_info, 
				sizeof (ci->hw_info) - 1);
	}
	ci->hw_info[sizeof (ci->hw_info) - 1] = 0;
	/* TODO */
	/* ci->intr_count = cards[cardnum]->intrcount;
	 * ci->ack_count = cards[cardnum]->ackcount;
	 */

	return 0;
}


void *alloc_mem(int nbytes) {
	return malloc(nbytes);
}


void free_mem(void *addr) {
	if (addr) {
		free(addr);
	}
}


int oss_disable_device(oss_device_t *osdev) {
	cmn_err(0, "not implemented: oss_disable_device()\n");
	return 0;
}


void oss_unregister_device(oss_device_t *osdev) {
	cmn_err(0, "not implemented: oss_unregister_device()\n");
}


int oss_register_device(oss_device_t *osdev, const char *name) {
	if ((osdev->name = PMALLOC (NULL, strlen (name) + 1)) == NULL) {
		cmn_err (CE_WARN, "Cannot allocate memory for device name.\n");
		osdev->name = "Unknown device";
	}
	strcpy (osdev->name, name);

	return OSS_ENXIO;
}


oss_device_t *osdev_clone(oss_device_t *orig_osdev, int new_instance) {
	oss_device_t *osdev;

	osdev = PMALLOC (NULL, sizeof (*osdev));
	if (osdev == NULL)
	{
		cmn_err (CE_WARN, "osdev_create: Out of memory\n");
		return NULL;
	}
	memcpy (osdev, orig_osdev, sizeof (*osdev));
	osdev->dev_type = DRV_CLONE;
	osdev->instance = new_instance;
	sprintf (osdev->nick, "%s%d", orig_osdev->modname, new_instance);
	sprintf (osdev->handle, "%s%d", orig_osdev->modname, new_instance);

	return osdev;
}


void oss_pci_byteswap(oss_device_t *osdev, int mode) {
	/* NOP */
}


void soundcard_attach(void) {
	memset(&core_osdev, 0, sizeof(oss_device_t));

#ifdef LICENSED_VERSION
	if (!oss_license_handle_time (oss_get_walltime ()))
	{
		cmn_err (CE_WARN, "This version of Open Sound System has expired\n");
		cmn_err (CE_CONT,
				"Please download the latest version from www.opensound.com\n");
		oss_expired = 1;
	}
#endif

	core_osdev.major = oss_major;
	oss_register_device (&core_osdev, "OSS core services");

	oss_common_init(&core_osdev);
}


oss_device_t *osdev_create(dev_info_t *dip, int dev_type, int instance, 
		const char *nick, const char *handle) {
	oss_device_t *osdev = NULL;
	int i; 

	if (handle == NULL)
		handle = nick;

	/*
	 * Don't accept any more drivers if expired
	 */
	if (oss_expired && oss_num_cards > 0)
		return NULL;
	for (i = 0; i < oss_num_cards; i++)
	{
		if (cards[i]->dip == dip)
		{
			osdev = cards[i];
			break;
		}
	}
	if (osdev == NULL)
	{
		if (oss_num_cards >= MAX_CARDS)
		{
			cmn_err (CE_WARN, "Too many OSS devices. At most %d permitted.\n",
					MAX_CARDS);
			return NULL;
		}

		if ((osdev = PMALLOC (NULL, sizeof (*osdev))) == NULL)
		{
			cmn_err (CE_WARN, "osdev_create: Out of memory\n");
			return NULL;
		}
		memset(osdev, 0, sizeof(*osdev));
		osdev->cardnum = oss_num_cards;
		cards[oss_num_cards++] = osdev;
	}
	osdev->dip = dip;
	osdev->osid = dip;
	osdev->instance = instance;
	osdev->available = 1;
	osdev->dev_type = dev_type;
	osdev->devc = NULL;
	osdev->first_mixer = -1;
	sprintf (osdev->nick, "%s%d", nick, instance);
	strcpy (osdev->modname, nick);

	/*
	 * Create the device handle
	 */
	switch (dev_type)
	{
		case DRV_PCI:
			{
				sprintf (osdev->handle, "OSS-PCI");
			}
			break;

		default:
			sprintf (osdev->handle, "%s%d", handle, instance);
	}

	return osdev;
}


static void init_fileinfo(struct fileinfo *fi, int flags, int pid) {
	memset (fi, 0, sizeof (*fi));
	if (flags & R_BIT) {
		fi->mode = OPEN_READ;
	}
	else if (flags & W_BIT) {
		fi->mode = OPEN_WRITE;
	}
	else if (flags & (R_BIT|W_BIT)) {
		fi->mode = OPEN_READWRITE;
	}
	else {
		cmn_err(CE_PANIC, "init_fileinfo: no modifier set\n");
	}
	fi->acc_flags = flags;
	fi->dev = -1;
	fi->pid = pid;
	fi->cmd = NULL;
}


int oss_open(int dev, int flags, int proc_nr) {
	oss_cdev_t *cdev;
	int retval;
	int tmpdev;

	if (dev >= oss_num_cdevs) {
		return ENXIO;
	}

	if ((cdev = oss_cdevs[dev]) == NULL || cdev->d == NULL) {
		return ENXIO;
	}

	if (cdev->d->open == NULL) {
		return ENODEV;
	}
	init_fileinfo(&files[dev], flags, proc_nr);
	tmpdev = dev;

	retval = cdev->d->open (cdev->instance, cdev->dev_class, 
			&files[dev], 0, 0, &tmpdev);

	if (tmpdev != -1) files[dev].dev = tmpdev;
	else files[dev].dev = dev;


	if (retval < 0) {
		return -retval;
	}

	open_devices++;
	return 0;
}


int oss_ioctl(int dev, int cmd, char *arg) {
	int retval;
	oss_cdev_t *cdev;

	if (dev >= oss_num_cdevs) {
		return ENXIO;
	}

	if ((cdev = oss_cdevs[dev]) == NULL || cdev->d == NULL) {
		return ENXIO;
	}

	if (cdev->d->ioctl == NULL) {
		return ENODEV;
	}

	retval = cdev->d->ioctl (cdev->instance, &files[dev], cmd, (ioctl_arg) arg);

	if (retval < 0)
		return -retval;
	return 0;
}


int oss_close(int dev) {
	oss_cdev_t *cdev;

	if (dev >= oss_num_cdevs)
		return ENXIO;

	if ((cdev = oss_cdevs[dev]) == NULL || cdev->d == NULL)
		return ENXIO;

	if (cdev->d->close == NULL)
	{
		return ENODEV;
	}

	cdev->d->close (cdev->instance, &files[dev]);
	open_devices--;

	return 0;
}


int oss_read(int dev, uio_t *buf) {
	oss_cdev_t *cdev;

	if (dev >= oss_num_cdevs) {
		return ENXIO;
	}

	if ((cdev = oss_cdevs[dev]) == NULL || cdev->d == NULL) {
		return ENXIO;
	}

	if (cdev->d->read == NULL) {
		return ENODEV;
	}

	return cdev->d->read (cdev->instance, &files[dev], buf, buf->uio_resid);
}


int oss_write(int dev, uio_t *buf) {
	oss_cdev_t *cdev;

	if (dev >= oss_num_cdevs) {
		return ENXIO;
	}

	if ((cdev = oss_cdevs[dev]) == NULL || cdev->d == NULL) {
		return ENXIO;
	}

	if (cdev->d->write == NULL) {
		return ENODEV;
	}

	return cdev->d->write (cdev->instance, &files[dev], buf, buf->uio_resid);
}

void oss_pcie_init(oss_device_t *osdev, int flags) {
	/* TODO: Should we do something? */
}


void oss_timer_callback(struct timer *tp) {
	void *arg;
	tmout_desc_t *tmout = (tmout_desc_t *) tmr_arg(tp)->ta_ptr;
	cmn_err(0, "oss_timer_callback\n");

	timeout_random++;

	if (!tmout->active) return;

	arg = tmout->arg;
	tmout->active = 0;
	tmout->timestamp = 0;

	tmout->func(arg);
}


#ifdef OSS_NO_LONG_LONG
timeout_id_t oss_timeout (void (*func) (void *), void *arg, unsigned long ticks) {
#else
timeout_id_t oss_timeout (void (*func) (void *), void *arg, unsigned long long ticks) {
#endif
	tmout_desc_t *tmout = NULL;
	int id, n;

	cmn_err(0, "oss_timeout()\n");
	timeout_random++;

	n = 0;
	id = -1;

	while (id == -1 && n < MAX_TMOUTS)
	{
		if (!tmouts[next_id].active)
		{
			tmouts[next_id].active = 1;
			id = next_id++;
			tmout = &tmouts[id];
			break;
		}
		next_id = (next_id + 1) % MAX_TMOUTS;
	}

	if (id == -1)         /* No timer slots available */
	{
		cmn_err (CE_WARN, "Timeout table full\n");
		return 0;
	}

	tmr_inittimer(&tmout->timer);
	tmout->func = func;
	tmout->arg = arg;
	tmout->timestamp = id | (timeout_random & ~0xff);

	set_timer(&tmout->timer, ticks, oss_timer_callback, tmout);

	return id | (timeout_random & ~0xff);
}



void oss_untimeout (timeout_id_t id) {
	tmout_desc_t *tmout;
	int ix;

	cmn_err(0, "oss_untimeout()\n");
	ix = id & 0xff;
	if (ix < 0 || ix >= MAX_TMOUTS)
		return;

	timeout_random++;
	tmout = &tmouts[ix];

	if (tmout->timestamp != id)   /* Expired timer */
		return;
	if (tmout->active)
		cancel_timer(&tmout->timer);
	tmout->active = 0;
	tmout->timestamp = 0;
}


	int
oss_find_minor (int dev_class, int instance)
{
	int i;

	for (i = 0; i < oss_num_cdevs; i++)
		if (oss_cdevs[i]->d != NULL && oss_cdevs[i]->dev_class == dev_class
				&& oss_cdevs[i]->instance == instance)
			return i;

	return -1;
}

